module LogicUS.FOL.Resolution exposing
    ( Resolvent, cfol2SeparationSubst, cfol2Separation, cfol2ContraryLiterals
    , ResolutionTableau, csfolSCFResolution
    , resolutionTableauToString, resolutionTableauToDOT
    , cfol2AllResolvents
    )

{-| The module provides the tools for aplying the differents resolution strategies to a set of propositional clauses for verifying its unfeasibility.


# Resolvents

@docs Resolvent, cfol2SeparationSubst, cfol2Separation, cfol2ContraryLiterals, cfol2AllResolvents.


# Refutationally Resolution Tableau Algorithm

@docs ResolutionTableau, csfolSCFResolution


# Resolution Tableau Representation

@docs resolutionTableauToString, resolutionTableauToDOT

-}

import Dict exposing (Dict)
import Graph exposing (Edge, Graph, Node)
import Graph.DOT as GDOT exposing (defaultStyles)
import List.Extra as LE
import LogicUS.AUX.AuxiliarFuctions exposing (uniqueConcatList)
import LogicUS.FOL.Clauses as FOL_CL exposing (ClauseFOL, ClauseFOLLiteral)
import LogicUS.FOL.SintaxSemantics as FOL_SS exposing (Substitution, Term(..))
import LogicUS.FOL.Unification as FOL_UN
import Set



-- Work functions


{-| It represent the information of the resolvent, the clauses involved, the renames done, the maximum general unifier (mgu) and the result of the resolvent
-}
type alias Resolvent =
    { c1 : ClauseFOL
    , c2 : ClauseFOL
    , r1 : Substitution
    , r2 : Substitution
    , mgu : Substitution
    , res : ClauseFOL
    }


type alias ResolutionItem =
    { c : ClauseFOL
    , p1 : Int
    , p2 : Int
    , r1 : Substitution
    , r2 : Substitution
    , mgu : Substitution
    }


{-| It represent the graph structure of the tableau
-}
type alias ResolutionTableau =
    Graph ( Bool, ClauseFOL ) ( Substitution, Substitution )


{-| It generates the appropriate substitutions for the separation of the clauses
-}
cfol2SeparationSubst : ClauseFOL -> ClauseFOL -> ( Substitution, Substitution )
cfol2SeparationSubst c1 c2 =
    let
        vs_r =
            Set.toList <|
                Set.intersect
                    (FOL_CL.cfolVarSymbols c1)
                    (FOL_CL.cfolVarSymbols c2)
    in
    let
        s1 =
            Dict.fromList <| List.map (\( s, is ) -> ( ( s, is ), Var ( s, is ++ [ 1 ] ) )) vs_r

        s2 =
            Dict.fromList <| List.map (\( s, is ) -> ( ( s, is ), Var ( s, is ++ [ 2 ] ) )) vs_r
    in
    ( s1, s2 )


{-| It generates the appropriate substitutions for the separation of the clauses
-}
cfol2Separation : ClauseFOL -> ClauseFOL -> ( ClauseFOL, ClauseFOL )
cfol2Separation c1 c2 =
    let
        ( s1, s2 ) =
            cfol2SeparationSubst c1 c2
    in
    ( FOL_CL.cfolApplySubstitution s1 c1, FOL_CL.cfolApplySubstitution s2 c2 )


{-| It searches all the pairs of literals from two clauses that have the same predicate and contrary sign
-}
cfol2ContraryLiterals : ClauseFOL -> ClauseFOL -> List ( ClauseFOLLiteral, ClauseFOLLiteral )
cfol2ContraryLiterals c1 c2 =
    let
        searchContraryLiteral ( a, sign ) c =
            List.filter (\( a_, sign_ ) -> FOL_CL.cfolAtomSymbol a == FOL_CL.cfolAtomSymbol a_ && sign /= sign_) c
    in
    List.foldl (\( a, sign ) ac -> ac ++ (List.map (\l -> ( ( a, sign ), l )) <| searchContraryLiteral ( a, sign ) c2)) [] c1


{-| It searches all the pairs of literals from two clauses that have the same predicate and contrary sign
-}
cfol2AllResolvents : ClauseFOL -> ClauseFOL -> List Resolvent
cfol2AllResolvents c1 c2 =
    let
        ( r1, r2 ) =
            cfol2SeparationSubst c1 c2

        c1_ =
            FOL_CL.cfolApplySubstitution r1 c1

        c2_ =
            FOL_CL.cfolApplySubstitution r2 c2
    in
    let
        cls =
            cfol2ContraryLiterals c1_ c2_
    in
    List.foldl
        (\( ( a1, sg1 ), ( a2, sg2 ) ) ac ->
            case FOL_UN.atomsMGU (FOL_CL.clauseFOLAtomToAtom a1) (FOL_CL.clauseFOLAtomToAtom a2) of
                Just mgu ->
                    let
                        c1__ =
                            List.filter (\x -> x /= ( FOL_CL.cfolAtomApplySubstitution mgu a1, sg1 )) <| FOL_CL.cfolApplySubstitution mgu c1_

                        c2__ =
                            List.filter (\x -> x /= ( FOL_CL.cfolAtomApplySubstitution mgu a2, sg2 )) <| FOL_CL.cfolApplySubstitution mgu c2_
                    in
                    ac ++ [ { c1 = c1, c2 = c2, r1 = r1, r2 = r2, mgu = mgu, res = FOL_CL.cfolUnion c1__ c2__ } ]

                Nothing ->
                    ac
        )
        []
        cls


recoverResolutionPath : Int -> Dict Int ResolutionItem -> ( List (Node ClauseFOL), List (Edge ( Substitution, Substitution )) )
recoverResolutionPath i refDict =
    case Dict.get i refDict of
        Just ri ->
            let
                ( nodes1, edges1 ) =
                    recoverResolutionPath ri.p1 refDict

                ( nodes2, edges2 ) =
                    recoverResolutionPath ri.p2 refDict
            in
            ( uniqueConcatList nodes1 nodes2 ++ [ Node i ri.c ]
            , uniqueConcatList edges1 edges2
                ++ [ Edge ri.p1 i ( ri.r1, Dict.filter (\k _ -> Set.member k (FOL_CL.cfolVarSymbols <| Maybe.withDefault [] <| Maybe.map .c <| Dict.get ri.p1 refDict)) ri.mgu )
                   , Edge ri.p2 i ( ri.r2, Dict.filter (\k _ -> Set.member k (FOL_CL.cfolVarSymbols <| Maybe.withDefault [] <| Maybe.map .c <| Dict.get ri.p2 refDict)) ri.mgu )
                   ]
            )

        _ ->
            ( [], [] )


resolventsWithClosedSCFResolutionAux : List ( Int, ResolutionItem ) -> Int -> ClauseFOL -> List ResolutionItem
resolventsWithClosedSCFResolutionAux closed id c =
    List.foldl
        (\( i, ri ) ac ->
            List.foldl
                (\r ac2 ->
                    if not <| FOL_CL.cfolIsTautology r.res || List.any (\x -> FOL_CL.cfolSubsumes x.c r.res) ac || List.any (\( _, x ) -> FOL_CL.cfolSubsumes x.c r.res) closed then
                        { c = r.res, p1 = id, p2 = i, r1 = r.r1, r2 = r.r2, mgu = r.mgu } :: List.filter (\x -> not <| FOL_CL.cfolSubsumes r.res x.c) ac2

                    else
                        ac2
                )
                ac
                (cfol2AllResolvents ri.c c)
        )
        []
        closed


openedUpdationSCFResolutionAux : List ( Int, ResolutionItem ) -> List ( Int, ResolutionItem ) -> List ( Int, ResolutionItem )
openedUpdationSCFResolutionAux xs new_opens =
    let
        res =
            List.foldl
                (\( li, ri ) ( ac, rest ) ->
                    let
                        add_ac =
                            LE.takeWhile (\( lx, _ ) -> lx <= li) rest
                    in
                    if List.any (\( _, x ) -> FOL_CL.cfolSubsumes x.c ri.c) (ac ++ add_ac) then
                        ( ac ++ add_ac, List.drop (List.length add_ac) rest )

                    else
                        ( ac ++ add_ac ++ [ ( li, ri ) ]
                        , List.filter (\( _, x ) -> not (FOL_CL.cfolSubsumes ri.c x.c)) <| List.drop (List.length add_ac) rest
                        )
                )
                ( [], xs )
                new_opens
    in
    Tuple.first res ++ Tuple.second res


csfolSCFResolutionAux : List ( Int, ResolutionItem ) -> List ( Int, ResolutionItem ) -> Int -> ( List (Node ClauseFOL), List (Edge ( Substitution, Substitution )) )
csfolSCFResolutionAux closed opened nid =
    case opened of
        [] ->
            ( [], [] )

        ( _, ri ) :: xs ->
            if List.isEmpty ri.c then
                let
                    refDict =
                        Dict.fromList <| closed ++ [ ( nid + 1, ri ) ]
                in
                recoverResolutionPath (nid + 1) refDict

            else
                let
                    r_closed =
                        resolventsWithClosedSCFResolutionAux closed (nid + 1) ri.c
                in
                let
                    new_closed =
                        closed ++ [ ( nid + 1, ri ) ]

                    new_opened =
                        openedUpdationSCFResolutionAux xs (List.sortBy (\x -> Tuple.first x) <| List.map (\x -> ( List.length x.c, x )) r_closed)
                in
                csfolSCFResolutionAux new_closed new_opened (nid + 1)


{-| It uses resolution algorithm using shorted clause first heuristic for determining the feasibilibity of a set of clauses. It gives the insatisfactibility and a graph with the resolution path to inconsitence. If clause set is feasible then a graph with only initial nodes is returned.

Note: You can render the graph with GraphViz Viewer and _resolutionTableauToDOt_ as we show in the example above.

-}
csfolSCFResolution : List ClauseFOL -> ( Bool, ResolutionTableau )
csfolSCFResolution clauses =
    let
        cs =
            FOL_CL.csfolRemoveEqClauses clauses
    in
    let
        new_cs =
            List.sortBy (\x -> Tuple.first x) <| List.map (\x -> ( List.length x, { c = x, p1 = 0, p2 = 0, r1 = Dict.empty, r2 = Dict.empty, mgu = Dict.empty } )) <| FOL_CL.csfolRemoveSubsumedClauses <| FOL_CL.csfolRemoveTautClauses <| cs
    in
    let
        ( nodes, edges ) =
            csfolSCFResolutionAux [] new_cs 0
    in
    let
        nid_max =
            Maybe.withDefault 0 <| List.maximum <| List.map (\x -> x.id) <| nodes

        nodes_clauses =
            List.map (\x -> x.label) <| nodes
    in
    let
        final_nodes =
            List.map (\x -> Node x.id ( List.member x.label cs, x.label )) nodes
                ++ (List.indexedMap (\i x -> Node (nid_max + i + 1) ( True, x )) <| List.filter (\x -> not (List.member x nodes_clauses)) cs)
    in
    ( edges /= [], Graph.fromNodesAndEdges final_nodes edges )



--     type alias ResolutionItem =
-- { c : ClauseFOL
-- , p1 : Int
-- , p2 : Int
-- , r1 : Substitution
-- , r2 : Substitution
-- , mgu : Substitution
-- }
-- type alias Resolvent =
-- { c1 : ClauseFOL
-- , c2 : ClauseFOL
-- , r1 : Substitution
-- , r2 : Substitution
-- , mgu : Substitution
-- , res : ClauseFOL
-- }


{-| Express a Resolution Tableau as a string.
-}
resolutionTableauToString : ResolutionTableau -> String
resolutionTableauToString g =
    let
        toStringNode =
            \( _, cs ) -> Just <| FOL_CL.cfolToString cs

        toStringEdge =
            \( rn, mgu ) ->
                if Dict.isEmpty rn then
                    Just <| FOL_SS.substitutionToString mgu

                else
                    Just <| FOL_SS.substitutionToString rn ++ "\n" ++ FOL_SS.substitutionToString mgu
    in
    Graph.toString toStringNode toStringEdge g


{-| Express a Resolution Tableau as a string in DOT format that is viewable with a GraphViz Render.
**Note:** If you are using elm repl, before introducing the code you must replace _\\n_ by _\\n_ and _\\"_ by _"_ in a simple text editor.
-}
resolutionTableauToDOT : ResolutionTableau -> String
resolutionTableauToDOT g =
    let
        myStyles =
            { defaultStyles | node = "shape=box, color=white, fontcolor=black", edge = "dir=none, color=blue, fontcolor=blue" }

        toStringNode =
            \( _, cs ) -> Just <| FOL_CL.cfolToString cs

        toStringEdge =
            \( rn, mgu ) ->
                if Dict.isEmpty rn && Dict.isEmpty mgu then
                    Nothing

                else if Dict.isEmpty rn then
                    Just <| FOL_SS.substitutionToString mgu

                else if Dict.isEmpty mgu then
                    Just <| FOL_SS.substitutionToString rn

                else
                    Just <| FOL_SS.substitutionToString rn ++ "<BR\\>" ++ FOL_SS.substitutionToString mgu

        initialNodes =
            String.join ";" <|
                List.map String.fromInt <|
                    List.foldl
                        (\x ac ->
                            if Tuple.first x.label then
                                ac ++ [ x.id ]

                            else
                                ac
                        )
                        []
                        (Graph.nodes g)
    in
    String.replace "\"" ">" <| String.replace "=\"" "=<" <| String.replace "\n}" ("\n\n  {rank=same; " ++ initialNodes ++ ";}\n}") <| GDOT.outputWithStyles myStyles toStringNode toStringEdge g
